/**
   
 */
/*  2000 Serge Rosmorduc,
 * quipe EC.art, Laboratoire LRIA, Universit de Paris 8
 *
 * 15 rue Catulienne
 * 93236 St Denis Cedex
 * FRANCE
 *
 * Tout droits rservs.
 * 
 * voyez le fichier ../LICENCE.txt pour plus d'informations
 * See the ../LICENCEE.txt file for more information
 *
 * $Id$
 *
 */

/*
 * $Log$
 */

#include "Bindings.h"
// #include <algo.h>

VarRep::VarRep(const string & vname, int pos)  
  : position(pos), name(vname)
{

}

const string &
VarRep::getName() const 
{
  return name;
}

Bindings::Bindings(const VarTable *t)
{
  tabNames= t;
  nbvars= t->getNbVars();
  tabValues= new const BoundExpr*[nbvars];
  fill(tabValues, tabValues+nbvars, (const BoundExpr *)0);
}

Bindings::Bindings(const Bindings& b)
{
  tabNames= b.tabNames;
  nbvars= b.nbvars;
  tabValues= new const BoundExpr* [nbvars];
  copy(b.tabValues, b.tabValues+b.nbvars, tabValues);
}

Bindings::~Bindings()
{
  delete [] tabValues;
}

void 
Bindings::set(const VarRep *v, const BoundExpr *value)
{
  tabValues[v->position]= value;
};

const BoundExpr *Bindings::get(const VarRep *v) const
{
  return   tabValues[v->position];
}

bool Bindings::hasvalue(const VarRep *v) const
{
  return (tabValues[v->position] != 0);
}

const VarRep *
Bindings::getRep(const string & varname) const {
  return tabNames->getVar(varname);
}

VarTable::VarTable()
{
}

VarTable::~VarTable()
{
  // map<string,VarRep *>::iterator i;
  // CE DESTRUCTEUR EST ON NE PEUT PLUS SUSPECT
  // for(i= rep.begin(); i != rep.end(); i++)
  // delete i->second;
}

VarRep *
VarTable::getVar(const string &vname)
{
  if (rep.find(vname) == rep.end())
    {
      int num= getNbVars();
      rep[vname]= new VarRep(vname, num);
    }
  return rep[vname];
}

const VarRep *
VarTable::getVar(const string &vname) const
{
  map<string, VarRep*>::const_iterator i;
  if ((i= rep.find(vname)) == rep.end())
    return 0;
  else
    {
      return i->second;
    }
}

int VarTable::getNbVars() const
{
  return rep.size();
}



