/**
   Classes used for variable representation and values are here.
 */
/*  2000 Serge Rosmorduc,
 * quipe EC.art, Laboratoire LRIA, Universit de Paris 8
 *
 * 15 rue Catulienne
 * 93236 St Denis Cedex
 * FRANCE
 *
 * Tout droits rservs.
 * 
 * voyez le fichier ../LICENCE.txt pour plus d'informations
 * See the ../LICENCEE.txt file for more information
 *
 * $Id$
 *
 */

/*
 * $Log$
 */

#ifndef BINDINGS_H
#define BINDINGS_H


class VarRep; // Internal representation of a variable
class Bindings; // links variable to their values
class VarTable; // links variable names to variable representation

#include"Expr.h"
#include <map>


class VarRep {
 public:
  VarRep(const string & vname, int pos);
  const string & getName() const; 
 private:
  int position;
  const string name;
  friend class Bindings;
};


class Bindings {
public:
  Bindings(const VarTable *t);
  Bindings(const Bindings& b);
  ~Bindings();
  // void set(const VarRep *v, const PlainExpr *value);
  void set(const VarRep *v, const BoundExpr *value);
  const BoundExpr *get(const VarRep *v) const;
  bool hasvalue(const VarRep *v) const;
  const VarRep *getRep(const string & varname) const;
private:
  // const PlainExpr **tabValues;
  const BoundExpr **tabValues;
  const VarTable *tabNames;
  int nbvars;
};

class VarTable {
 public:
  VarTable();
  ~VarTable();
  // getVar returns a VarRep for a given Variable name.
  // If the variable is a new one, a new VarRep is returned.
  // If the variable was already there, the old VarRep is returned.
  VarRep * getVar(const string & vname);
  // The const version doesn't create a new VarRep if none exists.
  // Instead, it sends back 0.
  // The methods' names should probably be different.
  const VarRep * getVar(const string & vname) const;
  /// Bindings * newBindings(); // Sends back a empty Bindings suitable for this rule. 
  int getNbVars() const;		// How many variables are there ?
 private:
  map<string,VarRep *> rep;
};


#endif



