/**
   
 */
/*  2000 Serge Rosmorduc,
 * quipe EC.art, Laboratoire LRIA, Universit de Paris 8
 *
 * 15 rue Catulienne
 * 93236 St Denis Cedex
 * FRANCE
 *
 * Tout droits rservs.
 * 
 * voyez le fichier ../LICENCE.txt pour plus d'informations
 * See the ../LICENCEE.txt file for more information
 *
 * $Id$
 *
 */

/*
 * $Log$
 */

#include "Expr.h"
// #include <algo.h>

#include <iostream>
using namespace std;

PlainExpr::PlainExpr() 
{
  //nextBindings= 0;
};

PlainExpr::~PlainExpr() {};

//  const Rule *
//  PlainExpr::getRule() {
//    return rule;
//  }

// La classe VarExpr

VarExpr::VarExpr(const VarRep *v)
{
  var= v;
  // var= r->getVarTable()->getVar(vname);
}

VarExpr::~VarExpr()
{
  // Nothing !
  // VarExpr is destroyed by ~VarTable
}

//  VarExpr::VarExpr(VarRep* v, Bindings *b) : Expr(b)
//  {
//    var= v;
//  }

//  VarExpr::VarExpr(const VarExpr& v, Rule *r): Expr(r)
//  {
//    var= v.var;
//  }

bool
VarExpr::match(const PlainExpr *e, Bindings *b1, Bindings *b2) const
{
  extern int debuglev;
  if (debuglev > 3)
    cout << "Well, let's try " << this->toString(b1) << " and " << e->toString(b2) << endl;
  return e->matchVar(this, b2, b1);
}

bool
VarExpr::matchNeg(const NegExpr *e, Bindings *b1, Bindings *b2) const
{
  if (hasvalue(b1))
    {
      return getVal(b1)->match(e, b2);
      // return ! getVal(b1)->match(e, b2, b1);
    }
  else
    {
      throw string("Variable non instancie et ngation");
      return true;
    }
}

bool
VarExpr::matchVar(const VarExpr *e, Bindings *b1, Bindings *b2) const
{
  // One of the two variables should have a value
  if (hasvalue(b1))
    return getVal(b1)->match(e, b2);
  else if (e->hasvalue(b2))
    return e->getVal(b2)->match(this, b1);    
  else
    throw string("Normalement Impossible");
  return false;
}

bool
VarExpr::matchFunc(const FuncExpr *e, Bindings *b1, Bindings *b2) const
{
  if (hasvalue(b1))
    {
      return getVal(b1)->match(e, b2);
    }
  else
    {
      BoundExpr *be;
      be= new BoundExpr(e, b2);
      setVal(be,b1);
    }
  return true;    
}

bool
VarExpr::hasvalue(const Bindings *b1) const
{
  if (b1)
    return b1->hasvalue(var);
  else
    return false;
}


string 
VarExpr::toString(const Bindings *b1) const
{
  if (hasvalue(b1))
    return b1->get(var)->toString();
  else
    return
      var->getName();
}

void 
VarExpr::setVar(const VarRep *v)
{
  var= v;
}

const VarRep *
VarExpr::getVar() const
{
  return var;
}

void 
VarExpr::setVal(const BoundExpr *v, Bindings *b) const
{
  b->set(var, v);
}

const BoundExpr *
VarExpr::getVal(const Bindings *b) const
{
  return b->get(var);
}


// FuncExpr

FuncExpr::FuncExpr(const string&  f, list<PlainExpr *> *e)
{
  functor= f;
  arg.setList(e);
}

FuncExpr::FuncExpr(const string&  f)
{
  functor= f;
  arg.setList(new list<PlainExpr *>);
}


FuncExpr::~FuncExpr()
{
}

bool
FuncExpr::match(const PlainExpr *e, Bindings *b1, Bindings *b2) const
{
  return e->matchFunc(this, b2, b1);
}

bool 
FuncExpr::matchVar(const VarExpr *v, Bindings *b1, Bindings *b2) const
{
  return v->matchFunc(this, b2, b1);
}

bool
FuncExpr::matchNeg(const NegExpr *e, Bindings *b1, Bindings *b2) const
{
  Bindings *b= 0;
  VarTable v;
  // cout << "trying not to match : " << this->toString(b1) << endl;
  b= new Bindings(*b2);
  bool res=(! e->getExpr()->match(this, b, b1));
  delete b;
  return res;
}

bool 
FuncExpr::matchFunc(const FuncExpr *v, Bindings *b1, Bindings *b2) const
{
  bool res;
  if (getFunctor() == v->getFunctor()
      && arg.size() == v->arg.size())
    {
      res= true;
      // All arguments should match.
      for (list<PlainExpr *>::iterator i= arg.begin(), j= v->arg.begin();
	   res && i != arg.end() && j != v->arg.end();
	   i++, j++)
	if (! (*i)->match(*j, b1, b2))
	  res= false;
    }
    else 
      res= false;
  return res;
}


string
FuncExpr::toString(const Bindings *b) const
{
  string s= getFunctor();
  if (getArgs()->size() != 0)
    {
      s+= "(";
      list<PlainExpr *>::const_iterator i;
      for (i=getArgs()->begin(); i !=getArgs()->end();)
	{
	  s+= (*i)->toString(b);
	  i++;
	  if (i != getArgs()->end())
	    s+= ", ";
	}

      s+= ")";
    }
  return s;
}


const string& 
FuncExpr::getFunctor() const
{
  return functor;
}

void 
FuncExpr::setFunctor(string s)
{
  functor= s;
}

list<PlainExpr *> const * 
FuncExpr::getArgs() const
{
  return arg.getList();
}

void 
FuncExpr::setArgs(list<PlainExpr *> *args)
{
  arg= args;
}

NegExpr::NegExpr(const FuncExpr *e) : exp(e)
{
}

NegExpr::~NegExpr()
{
  delete exp;
}

bool
NegExpr::match(const PlainExpr *v, Bindings *b1, Bindings *b2) const
{
  return v->matchNeg(this, b2, b1);
}

bool 
NegExpr::matchVar(const VarExpr *v, Bindings *b1, Bindings *b2) const
{
  // return v->match(this, b2, b1);
  return v->matchNeg(this, b2, b1);
}


bool NegExpr::matchNeg(const NegExpr *v, Bindings *b1, Bindings *b2) const
{
  throw string("We shouldn't try to match two negative item");
}

bool NegExpr::matchFunc(const FuncExpr *v, Bindings *b1, Bindings *b2) const
{
  return v->matchNeg(this, b2, b1);  
  // return v->match(this, b2, b1);
}


string 
NegExpr::toString(const Bindings *b) const
{
  return "! " +  getExpr()->toString(b);
}

const FuncExpr * NegExpr::getExpr() const
{
  return exp;
}

// EpsilonExpr

EpsilonExpr::EpsilonExpr()
{
}

EpsilonExpr::~EpsilonExpr()
{
}

bool
EpsilonExpr::match(const PlainExpr *v, Bindings *b1, Bindings *b2) const
{
  return false;
}

bool
EpsilonExpr::matchVar(const VarExpr *v, Bindings *b1, Bindings *b2) const
{
  return false;
}

bool
EpsilonExpr::matchNeg(const NegExpr *v, Bindings *b1, Bindings *b2) const
{
  return false;
}

bool
EpsilonExpr::matchFunc(const FuncExpr *v, Bindings *b1, Bindings *b2) const
{
  return false;
}

string
EpsilonExpr::toString(const Bindings *b) const
{
  return "epsilon";
}

AndExpr::AndExpr(list<PlainExpr *> * t) : terms(t)
{
}

AndExpr::~AndExpr() 
{
}

string 
AndExpr::toString(const Bindings *b) const
{
  string res;
  list<PlainExpr *>::const_iterator i;
  for(i= terms.begin(); i != terms.end(); i++)
    res+= (*i)->toString(b) + " & ";
  res.erase(res.length() - 3, 3);
  return res;
}

bool 
AndExpr::match(const PlainExpr *e, Bindings *b1, Bindings *b2) const
{
  bool res= true;
  list<PlainExpr *>::const_iterator i;
  //   if (terms.size() == 0)
  // return false;
  // 
  for(i= terms.begin(); res && i != terms.end(); i++)
    res &= (*i)->match(e,b1,b2);
  return res;
}

const list<PlainExpr *> * 
AndExpr::getTerms() const
{
  return terms.getList();
}


Expr::Expr(AndExpr *a) 
{
  rep= a;
}

Expr::Expr() 
{
  rep= new AndExpr(new list<PlainExpr *>());
}

Expr::~Expr()
{
  delete rep;
}

string 
Expr::toString(const Bindings *b) const {
  return rep->toString(b);
}

bool 
Expr::match(const PlainExpr *e, Bindings *b1, Bindings *b2) const
{
  return rep->match(e,b1,b2);
}

bool
Expr::isEpsilon() const {
  return rep->getTerms()->size() == 0;
}

const Expr * Expr::epsilonExpr= new Expr();

const Expr *
Expr::getEpsilon()
{
  return epsilonExpr;
}

ostream &
operator<<(ostream &s, const BoundExpr &e)
{
  s << e.toString();
  return s;
}

string
BoundExpr::toString() const
{
  if (expr)
    return getExpr()->toString(getBindings());
  else
    return "epsilon";
}

bool BoundExpr::match(const PlainExpr *v, Bindings *newBindings) const
{
  return expr->match(v, const_cast<Bindings*>(bindings), newBindings);
}

const BoundExpr *
BoundExpr::epsilonExpr= new BoundExpr(new EpsilonExpr(), 0);

const BoundExpr * 
BoundExpr::getEpsilon()
{
  return BoundExpr::epsilonExpr;
}
