/**
   
 */
/*  2000 Serge Rosmorduc,
 * quipe EC.art, Laboratoire LRIA, Universit de Paris 8
 *
 * 15 rue Catulienne
 * 93236 St Denis Cedex
 * FRANCE
 *
 * Tout droits rservs.
 * 
 * voyez le fichier ../LICENCE.txt pour plus d'informations
 * See the ../LICENCEE.txt file for more information
 *
 * $Id$
 *
 */

/*
 * $Log$
 */

#include "ExprTokeniser.h"
#include <cctype>
using namespace std;

// This class is used for token reading
// ID = [a-zA-Z][a-zA-Z0-9]*
// RULESEP = "->"
// OPENPAR = "("
// CLOSEPAR= ")"
// VAR = "$" [a-zA-Z0-9]+
// SLASH = "/"
// COMMA = ","
// spaces are omitted


ExprStreamTokeniser::ExprStreamTokeniser(istream& s) :
  input(s),   currentTokenValue(""), currentToken(NONE)
{
}


void
ExprStreamTokeniser::readID()
{
  char c;
  while ((input.get(c)) && isalnum(c) || c == '-' || c == '_')
    currentTokenValue+= c;
  if (input.good())
    input.putback(c);
}

void
ExprStreamTokeniser::readNum()
{
  char c;
  while ((input.get(c)) && isdigit(c))
    {
      currentTokenValue+= c;
      currentTokenNumValue= currentTokenNumValue*10 + c - '0';
    }
  if (input.good())
    input.putback(c);
}

enum ExprTokeniser::token
ExprStreamTokeniser::next()
{
  char c;
  currentTokenValue= "";
  currentToken= END;
  // Omit spaces and the like
  while ((input.get(c)) && isspace(c));
  if (input.good())
    {
      if (isdigit(c) || c == '-')
	{
	  bool neg;
	  if (c == '-')
	    {
	      neg= true;
	      currentTokenNumValue= 0;
	    }
	  else
	    {
	      neg= false;
	      currentTokenNumValue= c - '0';
	    }
	  currentTokenValue= c;
	  readNum();
	  if (neg)
	    currentTokenNumValue= -currentTokenNumValue;
	  currentToken= NUM;
	}
      else if (isalnum(c))
	{
	  currentToken= ID;
	  // We have an identifier. 
	  // Let's read it.
	  // Langage note :
	  //   The following line is safe and legal C++,
	  //   although the DECLARATION "string s= c" wouldn't be legal
	  currentTokenValue= c;
	  readID();
	  if (currentTokenValue == "epsilon")
	    currentToken= EPSILON;
	}
      else
	{
	  currentTokenValue= c;	  
	  switch(c)
	  {
	  case '$':	    
	    currentToken= VAR;
	    readID();
	    if (currentTokenValue == "")
	      currentToken= ERROR; // "$" is not allowed on its own.
	    break;
	  case '=':
	    input.get(c);
	    if (c == '>')
	      currentToken= RULESEP;
	    else
	      currentToken= ERROR;
	    break;
	  case '&':
	    currentToken= AND;
	    break;	    
	  case '!':
	    currentToken= NEG;
	    break;	    
	  case '|':
	    currentToken= OR;
	    break;	    
	  case '/':
	    currentToken= SLASH;
	    break;	    
	  case ',':
	    currentToken= COMMA;
	    break;
	  case '(':
	    currentToken= OPENPAR;
	    break;
	  case ')':
	    currentToken= CLOSEPAR;
	    break;
	  default:
	    currentToken= ERROR;
	  }
	}
    }
  return currentToken;
}

enum ExprTokeniser::token 
ExprStreamTokeniser::getToken() const
{
  return currentToken;
}

const string & 
ExprStreamTokeniser::getTokenValue() const
{
    return currentTokenValue;
}

int
ExprStreamTokeniser::getTokenNumValue() const
{
    return currentTokenNumValue;
}

ExprStreamTokeniser::~ExprStreamTokeniser()
{
}

ExprStringTokeniser::ExprStringTokeniser(const string &s) 
  : rep(s)
{
  fakeinput= new istrstream(rep.c_str());
  realworker= new ExprStreamTokeniser(*fakeinput);
};

enum ExprTokeniser::token ExprStringTokeniser::next()
{
  return realworker->next();
}


enum ExprTokeniser::token ExprStringTokeniser::getToken() const
{
  return realworker->getToken();
}


const string & 
ExprStringTokeniser::getTokenValue() const
{
  return realworker->getTokenValue();
}

int
ExprStringTokeniser::getTokenNumValue() const
{
  return realworker->getTokenNumValue();
}

ExprStringTokeniser::~ExprStringTokeniser()
{
  delete realworker;
  delete fakeinput;
}

