/**
   
 */
/*  2000 Serge Rosmorduc,
 * quipe EC.art, Laboratoire LRIA, Universit de Paris 8
 *
 * 15 rue Catulienne
 * 93236 St Denis Cedex
 * FRANCE
 *
 * Tout droits rservs.
 * 
 * voyez le fichier ../LICENCE.txt pour plus d'informations
 * See the ../LICENCEE.txt file for more information
 *
 * $Id$
 *
 */

/*
 * $Log$
 */

#ifndef EXPRTOKENISER_H
#define EXPRTOKENISER_H
#include <string>

// This class is used for token reading
// NUM = [0-9]+
// ID = [a-zA-Z0-9]+
// RULESEP = "->"
// OPENPAR = "("
// CLOSEPAR= ")"
// VAR = "$" [a-zA-Z0-9]+
// SLASH = "/"
// COMMA = ","
// spaces are omitted
//

#include <cctype>
using namespace std;

class ExprTokeniser {
public:
  enum token {NONE, ERROR, ID, NUM, VAR, AND, OR, NEG, RULESEP, END, EPSILON, OPENPAR=  '(', CLOSEPAR= ')', SLASH= '/', COMMA= ','};
  virtual enum token next() = 0; // Advances to next token
  virtual enum token getToken() const= 0; // get current token.
  virtual const string & getTokenValue() const = 0; // get the current token's value as a string
  virtual int  getTokenNumValue() const = 0;
  virtual ~ExprTokeniser() {};
};

class ExprStreamTokeniser : public ExprTokeniser {
 public:
  ExprStreamTokeniser(istream & s);
  enum ExprTokeniser::token next();
  enum token getToken() const;
  const string & getTokenValue() const;
  int  getTokenNumValue() const;
  ~ExprStreamTokeniser();
 private:
  istream &input;
  string currentTokenValue;
  int currentTokenNumValue;
  enum token currentToken;
  void readID();
  void readNum();
};



// ExprStringTokeniser could have inherited from ExprStreamTokeniser,
// as one can make a string into a stream, but this is an implementation
// inheritance, so if done, it shouldn't be public.
// The technical problem was that in the case where ExprStringTokeniser inherited from
// ExprStreamTokeniser, the stream (istrstream or isstream) wasn't created 
// At the correct point, and thus we had to adapt the original ExprStreamTokeniser class
// for this, although it didn't bring a much better organization.

// We had the choice between customizing ExprStreamTokeniser for 
// allowing the inheritance, or to do the whole stuff by hand. 


// istrstream is used instead of isstream, 
// which isn't available yet in g++

#include <strstream.h>

class ExprStringTokeniser : public ExprTokeniser {
 public:
  ExprStringTokeniser(const string &s);
  enum ExprTokeniser::token next();
  enum token getToken() const;
  const string & getTokenValue() const;
  int  getTokenNumValue() const;
  ~ExprStringTokeniser();
 private:
  string rep;
  ExprStreamTokeniser *realworker;
  istrstream *fakeinput;
};


#endif

