/**
   PointerList is a template that deal with lists of pointers used for
   building expressions.

   At creation time, or later, the list is given a value, which is a 
   list<> * template. Deletion of both this list and of the elements associated 
   with the pointers in the list is done by PointerList. Typical use is :

   class A {
   public:
     PointerList<toto> rep;
     A(list<toto *> *t) { rep.setList(t); };
     ~A() {}; // both t AND its elements are automatically freed 
   };
*/

/*  2000 Serge Rosmorduc,
 * quipe EC.art, Laboratoire LRIA, Universit de Paris 8
 *
 * 15 rue Catulienne
 * 93236 St Denis Cedex
 * FRANCE
 *
 * Tout droits rservs.
 * 
 * voyez le fichier ../LICENCE.txt pour plus d'informations
 * See the ../LICENCEE.txt file for more information
 *
 * $Id$
 *
 */

/*
 * $Log$
 */

#ifndef POINTLIST_H
#define POINTLIST_H
#include <list>

using namespace std;

template<class T> class PointerList {
 public:
  PointerList(list<T *> *t=0) : tab(t) {};
  ~PointerList() {
    freeList();
  };
  void setList(list<T *> *t) { freeList(); tab= t;}
  list<T *>::iterator begin()  const {return tab->begin();};
  list<T *>::iterator end()  const {return tab->end();};
  const list<T *> *getList() const {return tab;};
  int size() const {return tab->size();}
 private:
  void freeList() {
    if (tab) {
      list<T *>::iterator i;
      for (i= tab->begin(); i!= tab->end(); i++)
	delete *i;
      delete tab;    
    }
  };
  list<T *> *tab;
};

#endif


