/**
   
 */
/*  2000 Serge Rosmorduc,
 * quipe EC.art, Laboratoire LRIA, Universit de Paris 8
 *
 * 15 rue Catulienne
 * 93236 St Denis Cedex
 * FRANCE
 *
 * Tout droits rservs.
 * 
 * voyez le fichier ../LICENCE.txt pour plus d'informations
 * See the ../LICENCEE.txt file for more information
 *
 * $Id$
 *
 */

/*
 * $Log$
 */

#ifndef RULE_H
#define RULE_H

#include "Expr.h"
#include "PointerList.h"

class SimpleExprBuilder;
class RuleItem;

class Rule {
 public:
  ~Rule();
  const list<Expr *> * getLeft() const;
  const list<PlainExpr *> * getRight() const;

  int getCost() const;
  
  VarTable *getVarTable();
  const VarTable *getVarTable() const;
  
  Bindings * makeBindings() const;
  string toString(Bindings *b) const ;
  
  // Information about the rule
  // Left part is epsilon ???
  bool startsWithEpsilon() const;
  // Number of elements in left part of the rule
  int getNumLeft() const;
  // Number of elements in right part of the rule
  int getNumRight() const;
 private:
  PointerList<Expr> left;
  PointerList<PlainExpr> right;
  int cost;
  VarTable * varNames;
  // Computed variables
  int numLeft;
  int numRight;
    
  // Only ExprBuilder can build rules :
  Rule();
  void setLeft(list<Expr *> *);
  void setRight(list<PlainExpr *> *);
  void setCost(int c);
  friend class SimpleExprBuilder;
  friend class RuleItem;
};

class RuleItem {
 public:
  // Create a Rule Item corresponding to the beginning of a rule.
  RuleItem(const Rule * rule);
  // Duplicate a rule Item
  RuleItem(const RuleItem &ri);
  RuleItem & operator =(const RuleItem &ri);
  ~RuleItem();

  // can we call makeNext ?
  // operator bool();
  
  // makeNext(...) : If e can be matched, makeNext returns a new
  // RuleItem which corresponds to the rule after matching expression
  // e. If not, it returns 0.
  // BoundExpr & should be "const BoundExpr &",
  // But we are faced with an annoying problem :
  // We call the match() function from Expr, which 
  // might change the bindings, but won't 

  // BEWARE : TEMPORARY PATCH
  // Well, there's a problem with RuleItems and links.
  //

  RuleItem * makeMatch(const BoundExpr *);
  
  // Accessors :
  const Expr * getUpper() const;
  BoundExpr  * getLower() const;
  int getCost() const;
  Bindings * getBindings();
  
  // next : advances in the rule, regardless of matching anything.
  void next();

  // true if this RuleItem is the last for its rule
  bool isFinal(); 
  
  const Rule * getRule() const;

 private:
  // gives the low part its value
  void fixLow();
  // Instance variables
  //  position in rule
  const Rule *rule;		/* rule */
  list<Expr*>::const_iterator lpos; /* iterator for left part */
  list<PlainExpr*>::const_iterator rpos; /* iterator for right part  */
  enum RulePart {left, middle, right} part; /* part of the rule (see drawing above) */
  Bindings  *bindings;		/* bindings for the variables used in rules */
  // Auxiliary instance variables
  int itemNumber;		/* rank of item in left part */
  bool isFirst;			/* are we at the beginning of a rule ?? */
  BoundExpr *low;
  friend ostream & operator<<(ostream &, const RuleItem &);
};

// Actually, it would be more elegant to use a class with subclasses for the
// internal state of RuleItem. 
// But it would require dynamic allocation and a lot of code for a fairly simple case.

ostream & 
operator<<(ostream &s, const RuleItem &r);

#endif
