
#include "RuleTransducer.h"
#include "ExprBuilder.h"

// Modification :
// Rule links can point either to SyncRuleNode, or to MiddleRuleNode.
// Adding a rule between two SyncRuleNodes will correspond to :
// CurNode= begin
// For each item in the rule :
//      if there is already a link with this label
//         CurNode= end of this link
//      else
//         create a link and a node
// end for.
// This means that a link label should NOT 
// be linked to the corresponding rule (as there is no more one corresponding
// rule for a link)
// 
// So RuleItem should be different ?
// Are rule items needed in a node anyway in this case ?
// Variables could be internally called $1, ....
// Hence, no need for keeping their "real" name once the rule is built.
// 



RuleTransducer::RuleTransducer()
{
  start= makeNodeByLabel(0);
  end= 0;
}


RuleTransducer::~RuleTransducer()
{
}

RuleNode *
RuleTransducer::makeNodeByLabel(int l)
{
  RuleNode *res;
  res= new RuleNode(this, l);
  return res;
}

RuleNode *
RuleTransducer::getStart()
{
  return start;
}

RuleNode *
RuleTransducer::getEnd()
{
  return start;
}

void
RuleTransducer::setEnd(RuleNode * e)
{
  if (end == 0)
    end= e;
  else
    throw string("Error in Rule Transducer. Only one node should be final");
}


RuleLink* 
RuleNode::addLink(RuleNode * n1, string rule)
{
  RuleLink *res;
  SimpleExprBuilder parser;
  Rule *r= parser.makeRule(rule);
  res= new RuleLink(this,n1,r);
  links.push_back(res);
  return res;
}

bool 
RuleNode::isFinal()
{
  return final;
}

void
RuleNode::setFinal(bool f)
{
  final= f;
  if (final)
    automaton->setEnd(this);
}

RuleLinkList::RuleLinkList(RuleNode *n) :
    rep(n->links.begin()),
    end((n->links.end()))
{
}
 
bool
RuleLinkList::hasElements()
{
  return rep != end;
}

RuleLink *
RuleLinkList::getElement()
{
  return *rep;
}

void
RuleLinkList::next()
{
  rep++;
};

RuleNode::RuleNode(RuleTransducer *a, int lab)
{
  automaton= a;
  label=lab;
  final= false;
}

RuleLinkList 
RuleNode::makeRuleLinks()
{
  return RuleLinkList(this);
}

int
RuleNode::getLabel()
{
  return label;
}

RuleLink::RuleLink(RuleNode *n0, RuleNode *n1, Rule* r)
{
  start=n0;
  end= n1;
  rule=r;
  ri= new RuleItem(r);
}

Rule* 
RuleLink::getRule()
{
  return rule;
}

int
RuleLink::getCost() const
{
  return rule->getCost();
}

RuleItem * 
RuleLink::getRuleItem()
{
  return ri;
}


RuleNode* 
RuleLink::getEnd()
{
  return end;
}

