#ifndef RULETRANSDUCER_H
#define RULETRANSDUCER_H

// All about Rules
class RuleTransducer;
class RuleNode;
class RuleLinkList;
class RuleLink;

#include "Rule.h"

// A Rule-based transducer.
class RuleTransducer {
 public:
  RuleTransducer();
  ~RuleTransducer();
  RuleNode * getStart();
  RuleNode * getEnd();
  RuleNode * makeNodeByLabel(int l);
  void setEnd(RuleNode *);
 private:
  RuleNode *start;
  RuleNode *end;
};

// A node in the Rule Transducer ???
class RuleNode {
 public:
  RuleNode(RuleTransducer *r, int lab=0);
  RuleLinkList makeRuleLinks();
  int getLabel();
  void setLabel(int label);
  RuleLink* addLink(RuleNode *, string rule);
  bool isFinal();
  void setFinal(bool);
 private:
  RuleTransducer *automaton;
  friend class RuleLinkList;
  int label;
  list<RuleLink*> links;
  bool final;
};

// Links in the Rule transducer :
class RuleLink {
 public:
  RuleLink(RuleNode *n0, RuleNode *n1, Rule* r);
  Rule* getRule();
  int getCost() const;
  RuleItem * getRuleItem();
  RuleNode* getEnd();
 private:
  RuleNode *start, *end;
  Rule *rule;
  RuleItem *ri;
};

class RuleLinkList {
 public:
  RuleLinkList(RuleNode *r);
  bool hasElements();
  RuleLink *getElement() ;
  void next();
  RuleLinkList &operator++();
 private:
  list<RuleLink*>::iterator rep;
  list<RuleLink*>::iterator end;
};


#endif
