
#include "Transducer.h"
#include <iostream>
#include <algorithm> 
using namespace std;
/*
 * Sur les liens Composites :
 *	new CompositeLink(this,next, l1->getUp(), 0, 0);
 *		Pas de production. Le bas == epsilon
 *      CompositeLink *cl= new CompositeLink(this, next, 0, l2->getRuleItem(), 0);
 *		lien epsilon/epsilon
 *	new CompositeLink(this, 0, l1->getUp(), r, l2);
 *		lien "normal". r est le ritem, l2 est la fin de la rgle 
 *		(nouvelle classe drive de RItem, avec fin de rgle en sus ?)
 *	
 *	
 */

// For addLink. Could (should) be moved elsewhere.
#include "ExprBuilder.h"
#include <limits.h>
// Transducer

extern int debuglev;

Transducer::Transducer()
{
}

Transducer::~Transducer()
{
}

void
Transducer::computeShortestPath()
{
  // A) Topological Sort O(S+A)
  getStart()->depthFirst(sortedNodes);
  // B) Shortest path
  list<Node *>::iterator i= sortedNodes.begin();
  Node *first= *i;
  first->bestCost= 0;
  i++;
  // cout << "computing shortest path :" << endl;
  while (i != sortedNodes.end())
    {
      // LOOP INVARIANT : 
      //   we know the shortest path
      //   for all of X's prececessors
      // cout << "node is " << (**i) << "(" <<  (*i) << ")" << endl;
      LinkList *prevlist= (*i)->makePreviousLinks();
      //   Find N, the predecessor of X such that :
      //    N.bestCost() + cost of link l1 between N and X is minimal
      //   X.bestLink= l1;
      // C++ correct line :
      // (*i)->bestCost= numeric_limits<int>.max();
      // Currently usable line : 
      (*i)->bestCost= INT_MAX;
      while (prevlist->hasElements())
	{
	  Link * currLink= prevlist->getElement();
	  // cost for coming from currLink :
	  int cost=
	    currLink->getCost() + 
	    currLink->getStart()->bestCost;
	  // 
	  if ((*i)->bestCost > cost)
	    {
	      (*i)->bestCost= cost;
	      (*i)->bestLink= currLink;
	    }
	  prevlist->next();
	}
      delete prevlist;
      i++;
    }
  // C) Reverse the process and build the path
  // Look for *the* final node, let it be curr
  Node * curr= getEnd();
  if (curr == 0)
	throw string("No final node in this transducer");
  Link * bestLink= curr->bestLink;
  curr->bestLink= 0;
  while (curr != 0)
    {
      // cout << "Looking at node " << curr << "(";
      // cout << *curr << ")" << endl;
      if (bestLink)
	{
	  // We get the information
	  // cout << "best link is " << *bestLink << endl;
	  Node * prev= bestLink->getStart();
	  Link * nextBest= prev->bestLink;
	  // we put the link information in the starting node
	  prev->bestLink= bestLink;
	  // we retrieve the next link to consider
	  bestLink= nextBest;
	  // and we are ready for next node
	  curr= prev;
	}
      else 
	{
	  // We'd better be at the beginning :
	  if (curr != getStart())
	    throw string("No Path between start and end nodes");
	  else
	    {
	      // We stop
	      curr= 0;
	    }
	}
    }
  // Now, if there is a path, the lowest-cost path is indicated.
}


void
Transducer::showPath(ostream &o, int detail) const
{
  const Node *curr= getStart();
  int c=0;
  o << "Best path" << endl;
  while (curr)
    {
      if (curr->bestLink)
	{
	  if (detail > 0) 
	    {
	      if (curr->bestLink->getDown())
		o << *curr->bestLink->getDown() << " ";
	    }
	  else
	    o << *curr->bestLink << endl;
	  c+= curr->bestLink->getCost();
	  curr= curr->bestLink->getEnd();
	}
      else
	curr=0;
    }
  o << endl << "Total cost: " << c;
}


// Entry Transducer
EntryTransducer::EntryTransducer()
{
  // Create a 0 node.
  start= makeNodeByLabel(0);
  end= 0;
}

EntryTransducer::~EntryTransducer(void)
{
}

// Overloaded Methods

Node *
EntryTransducer::getStart()
{
  return start;
}

const Node *
EntryTransducer::getStart() const
{
  return start;
}

Node *
EntryTransducer::getEnd()
{
  return end;
}

Node *
EntryTransducer::getCurrentNode()
{
  return currentNode->second;
}


void
EntryTransducer::nextNode()
{
  currentNode++;
}

void
EntryTransducer::reset()
{
  currentNode= nodes.begin();
}


bool
EntryTransducer::lastNode()
{
  return (currentNode== nodes.end());
}

// "Real" access functions :
EntryNode * 
EntryTransducer::getTrueStart()
{
  return start;
}

void
EntryTransducer::setEnd(EntryNode *e)
{
  if (end == 0)
    end= e;
  else
    throw string("Error in EntryTransducer. Only one node should be final");
}

EntryNode * 
EntryTransducer::getTrueCurrentNode()
{ 
  return currentNode->second; 
}

// Specific Methods

EntryNode *
EntryTransducer::makeNodeByLabel(int l)
{
  EntryNode *res;
  res= new EntryNode(this, l);
  nodes[l]= res;
  return res;
}

// Composite Transducer

CompositeTransducer::CompositeTransducer(Transducer* a1, RuleTransducer* a2)
{
  // The last part of the label is either :
  //   * a rule node label
  //   * a dynamically created label
  // Hence, maxLabel must be greater than the maximum label number in 
  //   the RuleTransducer a2.
  // maxLabel= a2->getMaxLabel() + 1; 
  // idea :
  
  // create first node.
  // push it on todo
  maxLabel= 0;
  start= new ChoiceNode(this, a1->getStart(), a2->getStart());
  end= 0;
  while (! toDo.empty())
    {
      CompositeNode *n;
      n= toDo.top();
      toDo.pop();
      if (debuglev > 1)
	cout << "extending node " << *n << endl;
      n->extends();
      if (n->isDeadNode())
	{
	  delete n;
	}
      if (debuglev > 10)
	cout << *this << endl;
    }
}

CompositeTransducer::~CompositeTransducer()
{
}

Node * 
CompositeTransducer::getStart()
{
  return start; 
}

const Node * 
CompositeTransducer::getStart() const
{
  return start; 
}

Node * 
CompositeTransducer::getEnd()
{
  return end; 
}

void
CompositeTransducer::setEnd(CompositeNode * e)
{
  if (end == 0)
    end= e;
  else
    throw string("Error in CompositeTransducer. Only one node should be final");
}


Node * 
CompositeTransducer::getCurrentNode()
{
    return currentNode->second;
}

void 
CompositeTransducer::nextNode()
{
  currentNode++;
}

void 
CompositeTransducer::reset()
{
  currentNode= nodes.begin();
}

bool 
CompositeTransducer::lastNode()
{
  return (currentNode== nodes.end());
}

// "Real" access functions :
CompositeNode * 
CompositeTransducer::getTrueStart()
{
  return start;
}

CompositeNode * 
CompositeTransducer::getTrueCurrentNode()
{ 
  return currentNode->second; 
}
  
// Specific Methods

ChoiceNode*
CompositeTransducer::getChoiceNode(Node *n, RuleNode *r)
{
  ChoiceNode *res;
  pair<Node *,RuleNode *> p(n,r);
  Labels::iterator pt;
  if (debuglev > 3)
    {
      cout << "looking for node for : " << endl;
      cout << "entry node : " << *n << endl;
      cout << "Rule node : for" << r << endl;
    }
  pt= labelValues.find(p);
  if (pt != labelValues.end())
    {
      // The node was found
      res= pt->second;
    }
  else
    {
      // Create a new node.
      res= new ChoiceNode(this,n,r);
      // It should be treated one day
      // toDo.push(res);
      // Register it as the son of n and r.
      labelValues[p]= res;
    }
  return res;
}

int
CompositeTransducer::nextLabel()
{
  return maxLabel++;
}

int
CompositeTransducer::getMaxLabel()
{
  return maxLabel;
}

void
CompositeTransducer::addToAgenda(CompositeNode *n)
{
  toDo.push(n);
}


// Base Node Class

Node::Node()
{
  bestLink= 0;
  state= white;
  bestCost= -1;
}

Node::~Node()
{
}

void
Node::depthFirst(list<Node*> &l)
{
  // Depth First Graph Walk from Cormen/Leiserson/Rivest
  state= gray;
  LinkList * adj= makeLinks();
  while (adj->hasElements())
    {
      Link * link= adj->getElement();
      Node * node= link->getEnd();
      if (node->state == white)
	{
	  node->depthFirst(l);
	}
      adj->next();
    }
  delete adj;
  state= black;
  // We add 
  l.push_front(this);
}

// EntryNode

EntryNode::EntryNode(EntryTransducer *a, int l)
{
  final= false;
  automaton= a;
  label= l;
}

EntryNode::~EntryNode()
{
}

// Inherited functions

bool 
EntryNode::isFinal()
{
  return final;
}


Transducer * 
EntryNode::getTransducer()
{
  return automaton;
}


int
EntryNode::getLabel()
{
  return label;
}

LinkList *
EntryNode::makeLinks()
{
  return new LinkListFromEntry(this, forward);
}

LinkList *
EntryNode::makePreviousLinks()
{
  return new LinkListFromEntry(this, backward);
}

// Specific functions

void 
EntryNode::setFinal(bool f)
{
  final= f;
  if (final)
    {
      automaton->setEnd(this);
    }
}

void
EntryNode::setLabel(int l)
{
  label=l;
}

EntryLink * 
EntryNode::addLink(EntryNode *end, 
		   const string & up,
		   const string & down,
		   int cost)
{
  EntryLink *res;
  BoundExpr *bup, *bdown;
  SimpleExprBuilder parse;
  // Create Bounded expressions.
  bup= new BoundExpr(parse.makePlainExpr(up), 0);
  bdown= new BoundExpr(parse.makePlainExpr(down), 0);
  res= new EntryLink(this, end, bup, bdown, cost);
  links.push_back(res);
  end->previousLinks.push_back(res);
  
  return res;
}

// CompositeNode

CompositeNode::CompositeNode(CompositeTransducer *a, Node *e)
{
  entry= e;
  a->addToAgenda(this);
  automaton= a;
  label= a->nextLabel();
  // The following is NOT NICE. UGLY BOOOOOH
  // (it registers the node in the list of nodes)
  a->nodes[label]= this;
}

CompositeNode::~CompositeNode()
{
  automaton->nodes.erase(automaton->nodes.find(label));
}


// Inherited functions

Transducer * 
CompositeNode::getTransducer()
{
  return automaton;
}

int
CompositeNode::getLabel()
{
  return label;
}

LinkList *
CompositeNode::makeLinks()
{
  return new LinkListFromComposite(this, forward);
}

LinkList *
CompositeNode::makePreviousLinks()
{
  return new LinkListFromComposite(this, backward);
}

// Specific functions


Node *
CompositeNode::getUpNode()
{
  return entry;
}


CompositeTransducer *
CompositeNode::getCompAutomaton()
{
  return automaton;
}

CompositeLink*
CompositeNode::addLink(CompositeNode *n1,
		       const BoundExpr *u,	/* upper label */
		       const RuleItem *d, /* Rule item. determine lower label */
		       int cost)	
{
  return new CompositeLink(this,n1,u,d, cost);
}

void
CompositeNode::registerLink(CompositeLink * l, enum LinkDir d)
{
  list<CompositeLink*> *mylistp;
  if (d== forward)
    mylistp= &links;
  else
    mylistp= &previousLinks;

  mylistp->push_back(l);
}

void
CompositeNode::unRegisterLink(CompositeLink * l, enum LinkDir d)
{
  list<CompositeLink*> *mylistp;
  if (d== forward)
    mylistp= &links;
  else
    mylistp= &previousLinks;

  list<CompositeLink*>::iterator elt;
  elt= find(mylistp->begin(), mylistp->end(), l);
  if (elt != mylistp->end())
    {
      mylistp->erase(elt);
    }
  else
    {
      throw string("should not happen : unregisterLink should always succeed !");
    }
}

// ChoiceNode

ChoiceNode::ChoiceNode(CompositeTransducer *a, Node *n, RuleNode *r) :
  CompositeNode(a,n), ruleNode(r)
{
  final= n->isFinal() && r->isFinal();
  if (final)
    {
      getCompAutomaton()->setEnd(this);
    }
}

ChoiceNode::~ChoiceNode()
{
}

bool
ChoiceNode::isFinal()
{
  return final;
}

RuleNode*
ChoiceNode::getRuleNode()
{
  return ruleNode;
}

// ChoiceNodes never die
bool
ChoiceNode::isDeadNode()
{
  return false;
}

void
ChoiceNode::extends()
{
  // Le noeud est le noeud (na,nb)
  Node *ANode= getUpNode();
  RuleNode *BNode= getRuleNode();
  
  // D'abord, les Epsilon-arcs partant de ANode :
  LinkList *l1= ANode->makeLinks();
  while (l1->hasElements()) {
    // si arc1 est un Epsilon-arc
    if (l1->getElement()->downIsEpsilon())
      {
	// Find the node (creating it if needed) :
	ChoiceNode* next;
	next= getCompAutomaton()->getChoiceNode(l1->getElement()->getEnd(), BNode);
	//    crer un arc vers le noeud. L'arc est un arc "normal"
	//     (il n'a pas d'information  porter)
	//    Est-ce que ce doit tre un CompositeLink ??? Non en fait. 
	CompositeLink* nlink= addLink(next, l1->getElement()->getUp(), 0, 
				      l1->getElement()->getCost());
	if (debuglev > 2)
	  {
	    cout << "adding link " << *nlink << " due to epsilon entry." << endl;
	  }
      }
    l1->next();
  }
  delete l1;
  // Ensuite, les Epsilon-arcs partant de BNode :
  RuleLinkList l2= BNode->makeRuleLinks();
  for (; l2.hasElements(); l2.next())
    {
      // Si la rgle commence par Epsilon
      if (l2.getElement()->getRule()->startsWithEpsilon())
	{
	  //      Crer le noeud (na,N++)
	  //      crer le CompositeLink correspondant (Epsilon/Epsilon)
	  int cost;
	  CompositeNode *next;
	  RuleItem *ri= new RuleItem(*l2.getElement()->getRuleItem());
	  if (ri->isFinal())
	    {
	      next= getCompAutomaton()->getChoiceNode(ANode,
						      l2.getElement()->getEnd());
	      cost= ri->getCost();
	    }
	  else
	    {
	      RuleItem *nextRI= new RuleItem(*ri);
	      nextRI->next();
	      next= new MiddleNode(getCompAutomaton(),
				   ANode,
				   nextRI,
				   l2.getElement());
	      cost= 0;		// Le calcul est plus complexe en fait.
	    }
	  CompositeLink* nlink= addLink(next, 0, ri, cost);
	  if (debuglev > 2)
	    {
	      cout << "adding link " << *nlink << " due to " << l2.getElement()->getRule()->toString(0) << endl;
	    }
	}
    }
  // Pour tout les autres arcs arc1,arc2
  l1= ANode->makeLinks();
  while (l1->hasElements()) {
    if (! l1->getElement()->downIsEpsilon())	
      for (RuleLinkList l2= BNode->makeRuleLinks(); l2.hasElements(); l2.next())
	{
	  RuleItem * r;
	  if (debuglev > 3)
	    {
	      cout << "trying " << *l2.getElement()->getRuleItem() << " ";
	      cout << "with " << *l1->getElement() << endl;
	    }
	  // Si la rgle ne commence pas par Epsilon 
	  if (! l2.getElement()->getRule()->startsWithEpsilon() 
	      && (r=  l2.getElement()->getRuleItem()->makeMatch(l1->getElement()->getDown())))
	    {
	      //    si a1.down() matche -> ritem r
	      //    crer le rarc avec r + bonne fin de rgle
	      if (debuglev > 1)
		{
		  cout << "matched " << *l2.getElement()->getRuleItem() << " ";
		  cout << "with " << *l1->getElement() << endl;
		}

	      CompositeNode *next;
	      int cost= r->getCost() + l1->getElement()->getCost();
	      if (r->isFinal())
		{
		  // crer SI BESOIN le ChoiceNode nfin(a1),nfin(a2).
		  next= getCompAutomaton()->getChoiceNode(l1->getElement()->getEnd(),
							  l2.getElement()->getEnd());
		  // delete r; Pourquoi ??????
		}
	      else
		{
		  // Sinon
		  // crer le noeud nfin(a1),n++
		  RuleItem *nextRI= new RuleItem(*r);
		  // we go to next rule Item.
		  nextRI->next();
		  next= new MiddleNode(getCompAutomaton(),
				       l1->getElement()->getEnd(),
				       nextRI,
				       l2.getElement()
				       );
		}
	      if (debuglev > 1)
		cout << *r << "cost " << cost << endl;
	      // CompositeLink *nlink= 
	      // addLink(next, l1->getElement()->getUp(), l2->getRuleItem(), cost)
	      CompositeLink *nlink= 
		addLink(next, l1->getElement()->getUp(), r, cost);
	      if (debuglev > 2)
		{
		  cout << "adding link " << *nlink << " due to rule match" << endl;
		}

	      //    Relier, c'est prt
	    }
	}
    l1->next();
  }
  delete l1;
  //  Fin pour
}

// MiddleNode

MiddleNode::MiddleNode(CompositeTransducer *a, // the transducer
		       Node * e,		// the entry node
		       RuleItem *r,	// the rule state
		       RuleLink *rl)	// the associated rule link
  :
CompositeNode(a,e), ruleItem(r), ruleLink(rl)
{
 
}


MiddleNode::~MiddleNode()
{
}

bool
MiddleNode::isFinal()
{
  return false;
}

bool
MiddleNode::isDeadNode()
{
  bool res= false;
  // Due to the way it is built, a MiddleNode has
  // Only one backward link.
  CompositeLink * l= (*previousLinks.begin());
  CompositeNode * prev= l->start;
  // The reason the previous line is ugly (use rather private part of
  // data external to MiddleNode) is that the system architecture supposes
  // that Transducers are manipulated in a generic way. The problem is that here
  // we are doing specific stuff.
  if (links.size() == 0) // No next link at all
    {
      res= true;
      // unlink the node
      delete l;
      // prev->unRegisterLink(l, forward);
      // this->unRegisterLink(l, backward);
      // The previous node might also be dead
      if (prev->isDeadNode())
	{
	  delete prev;
	}
    }
  return res;
}

RuleItem*
MiddleNode::getRuleItem()
{
  return ruleItem;
}

RuleLink*
MiddleNode::getRuleLink()
{
  return ruleLink;
}

void 
MiddleNode::extends()
{
  // Le noeud est le noeud (na,nb)
  Node *ANode= getUpNode();

  RuleItem *r= getRuleItem();	// r is a copy of the RITEM associated with the node
  // r.next();

  // First, let's work with epsilon links in the entry.
  LinkList *l1= ANode->makeLinks();
  while (l1->hasElements()) {
    if (l1->getElement()->downIsEpsilon())
      {
	// The new node corresponds to :
	//   * same state as far as rules are concerned
	//   * the corresponding node in the entry automaton
	//     is l1's end node.

	// Note : the ruleItem is shared between this node and
	//        the new node.
	CompositeNode * next;
	next= new MiddleNode(getCompAutomaton(),
			     l1->getElement()->getEnd(),
			     getRuleItem(),
			     getRuleLink());
	int cost= l1->getElement()->getCost();
	CompositeLink *nlink= addLink(next, l1->getElement()->getUp(), 0, cost);
	if (debuglev > 2)
	  {
	    cout << "adding link " << *nlink << " due to epsilon entry " << *l1->getElement()<< endl;
	  }
      }
    l1->next();
  }
  delete l1;
  // Below, we see if the rule part is an epsilon node

  if (r->getUpper()->isEpsilon()) // Epsilon-arc
    {
      // Two cases : either we are at rule's end, or not.
      // RuleItem *ri= new RuleItem(r);
      CompositeNode *next;
      if (r->isFinal())
	{
	  next= getCompAutomaton()->getChoiceNode(ANode,
						  getRuleLink()->getEnd());
	}
      else
	{
	  RuleItem *nextRI= new RuleItem(*r);
	  nextRI->next();
	  next= new MiddleNode(getCompAutomaton(),
			       ANode,
			       nextRI,
			       getRuleLink());
	}
      // No upper entry
      int cost= r->getCost();
      CompositeLink *nlink= addLink(next, 0, r, cost);
      if (debuglev > 2)
	{
	  cout << "adding link " << *nlink << " due to epsilon rule " << *r << endl;
	}
    }
  else
    {
      LinkList *l1= ANode->makeLinks();
      while (l1->hasElements()) {
	if ( ! l1->getElement()->downIsEpsilon())
	  {
	    // does l1 fit ?
	    int cost= l1->getElement()->getCost() + r->getCost();	    
	    if (RuleItem *newri=  r->makeMatch(l1->getElement()->getDown()))
	      {
		CompositeNode *next;
		// First, we create the link.
		if (newri->isFinal())
		  {
		    // Synchronization point between rules :
		    //  let's create a ChoiceNode
		    next= getCompAutomaton()->getChoiceNode(l1->getElement()->getEnd(), getRuleLink()->getEnd());
		  }
		else
		  {
		    RuleItem *nextRI= new RuleItem(*newri);
		    nextRI->next();
		    // We are still in the middle of the rule.
		    next= new MiddleNode(getCompAutomaton(),
					 l1->getElement()->getEnd(),
					 nextRI,
					 getRuleLink());
		  }
		CompositeLink* nlink= addLink(next, l1->getElement()->getUp(), newri, cost);
		if (debuglev > 2)
		  {
		    cout << "adding link " << *nlink << " due to " << r <<
		      " and " << *l1->getElement() << endl;
		  }
	      }
	  }
	l1->next();
      }
      delete l1;
    }
}


// Link

Link::Link()
{
}

Link::~Link()
{
}

// EntryLink
EntryLink::EntryLink(EntryNode *n1, EntryNode* n2, BoundExpr *u,
		     BoundExpr *d, int c)
{
  start=n1;
  end=n2;
  up= u;
  down= d;
  cost= c;
}

// Inherited Methods

Node *
EntryLink::getStart()
{
  return start;
}


Node *
EntryLink::getEnd()
{
  return end;
}

const BoundExpr *
EntryLink::getUp() const
{
  return up;
}

const BoundExpr *
EntryLink::getDown() const
{
  return down;
}

int
EntryLink::getCost() const
{
  return cost;
}

bool 
EntryLink::downIsEpsilon() const
{
  return down == 0 || down->getExpr()->isEpsilon();
}

// CompositeLink

CompositeLink::CompositeLink(CompositeNode *s, CompositeNode *e, 
			     const BoundExpr *u, const RuleItem *d, int c)
{
  if (u==0)
    u= BoundExpr::getEpsilon();
  up= u;
  down= d;
  start=s;
  end= e;
  cost= c;

  // Register the links
  s->registerLink(this, forward);
  e->registerLink(this, backward);
}

CompositeLink::~CompositeLink()
{
  // unRegister the links
  cost = 11111111;
  start->unRegisterLink(this, forward);
  end->unRegisterLink(this, backward);
}

// Inherited Methods
Node * 
CompositeLink::getStart()
{
  return start;
}

Node * 
CompositeLink::getEnd()
{
  return end;
}

const BoundExpr *
CompositeLink::getUp() const
{
  return up;
}

const BoundExpr *
CompositeLink::getDown() const
{
  if (down)
    return down->getLower();
  else
    return 0;
}

int
CompositeLink::getCost() const
{
  return cost;
}

bool 
CompositeLink::downIsEpsilon() const
{
  return
    (down == 0) || 
    (down->getLower() == 0) ||
    (down->getLower()->getExpr() == 0) ||    
    (down->getLower()->getExpr()->isEpsilon());
}


//  RuleLink *
//  CompositeLink::getRuleLink()
//  {
//    return ruleLink;
//  }


// I/O Facilities

ostream &
operator<<(ostream &s, Transducer &a)
{
  a.reset();
  while (! a.lastNode())
    {
      Node *n= a.getCurrentNode();
      s << *n << endl;
      LinkList* l= n->makeLinks();
      while(l->hasElements())
	{
	  s << "\t" << *l->getElement() << endl;
	  l->next();
	}
      delete l;
      a.nextNode();
    }
  return s;
}

ostream &
operator<<(ostream &s, Node &nod)
{
  s << "Node " << nod.getLabel();
  if (nod.isFinal())
    s << " (final)";
  return s;
}

ostream &
operator<<(ostream &s, Link &l)
{
  s << "Link " ;
  s << *l.getStart() << " -" ;
  s << *l.getUp()  << " : " ;
  if (l.getDown())
    s << *l.getDown() << "- " ;
  else
    s << "epsilon - " ;    
  s << *l.getEnd();
  s << " cost: " << l.getCost();
  return s;
}

/*
  Algorithme de calcul des solutions optimales :

  A) Ordonner les noeuds. Nous avons un graphe ordonn acyclique 
  cause du premier transducteur (vrifier cette assertion....)
  

  B) Empiler les noeuds ordonns.
  
  C) prendre le noeud du sommet de la pile.

  Invariant : tous les noeuds qui le prcdent ont t traits.

  Calculer le plus court chemin  rebours :
  
  min(N)= + infini
  Pour tout les arc a qui mnent  N
    Soit N0 le noeud de dpart de a.
    Si min(N0) + cot(a) < min(N)
    alors 
      min(N)= cot(a) + min(N0)
      prev(N)= a
    fin si
  finpour

  Pour avoir les k meilleurs chemins, ont fait la mme chose avec 
  un tableau de k arcs et k cots.n

  Quand on a fini, on n'a plus qu' partir du noeud final :
  N= noeud_final
  tant que N <> noeud_initial
    a= prev(N)
    N0= depart de a
    suiv(N0)= a
    N= N0
  fintantque
  
*/
