//
// Transliteration Program
// Version 0.1
// temporary before going TCL and clean
//
// Usage : translit signs.txt rules.txt < text.txt
//
// Where
//	text.txt : text to transliterate
//	signs.txt : rules for sign value
//	rules.txt : sign combination rules


#include "Transducer.h"
#include <iostream>
#include <fstream>
using namespace std;

RuleTransducer *
makeRuleTransducerFromFile(const string &fname);

RuleTransducer * 
makeInsertSep();

RuleTransducer *
makeWordCost();

int trace= 0;
int debuglev= 0;

int
main(int argc, char ** argv)
{
  cout << "taille lien" << sizeof(CompositeLink) << endl;
  cout << "taille noeuds" << sizeof(ChoiceNode) << endl; 
  cout << "taille noeuds" << sizeof(MiddleNode) << endl; 
  cout << "taille RuleItem" << sizeof(RuleItem) << endl; 
  cout << "taille BoundExpr" << sizeof(BoundExpr) << endl; 
  cout << "taille Bindings" << sizeof(Bindings) << endl; 
  
  istream *tmp;
  if (argc != 6 && argc != 7)
    {
      cerr << "usage: " << argv[0] << " codes.txt signs.txt rules.txt [erules.txt]" << endl;
      cerr << " where" << endl;
      cerr << "\t" << "codes.txt : codes for signs" << endl;
      cerr << "\t" << "signs.txt : rules for sign value" << endl;
      cerr << "\t" << "rA.txt : first combination rules" << endl;
      cerr << "\t" << "rB.txt : second combination rules" << endl;
      cerr << "\t" << "rC.txt : third combination rules" << endl;
      cerr << "the text to transliterate is on the standard input" << endl;
      exit(1);
    }

  try {
    // 
    // Create an Entry Transducer from standard input
    //
    
    if (argc== 6)
    	tmp= &cin;
    else
    	tmp= new ifstream(argv[6]);
  	
    EntryTransducer a0;
    EntryNode * n0= a0.getTrueStart();
    EntryNode * n1;
    string entry;
    int labnum= 1;
    while (*tmp >> entry)
      {
	n1= a0.makeNodeByLabel(labnum++);
	n0->addLink(n1, entry, entry);
	n0= n1;
      }
    n0->setFinal(true);
    cout << "first automaton" << endl;
    if (debuglev) cout << a0 << endl;

	if (argc==7) delete tmp;
	
    //
    // Normalize sign values
    //
    
    cout << "Normalize" << endl;
    RuleTransducer *normalize= makeRuleTransducerFromFile(argv[1]);
    CompositeTransducer a1(&a0,normalize);

    //    sleep(200);

    if (debuglev) cout << a1 << endl;

    //
    // Insert separators between entries
    //
    // cout << "Insert separators between entries" << endl;
    // RuleTransducer * insertSep= makeInsertSep();
    // CompositeTransducer a2(&a1,insertSep);

    // if (debuglev) cout << a2 << endl;    

    //
    // Insert sign values
    //
    cout << "Insert signs values" << endl;
    RuleTransducer *signValues= makeRuleTransducerFromFile(argv[2]);
    CompositeTransducer a3(&a1,signValues);

    if (debuglev) cout << a3 << endl;    

    // a3.computeShortestPath();
    // a3.showPath(cout);
    
    //debuglev=3;
    //
    // Apply Composition rules
    //
    cout << "Rules 1" << endl;
    RuleTransducer *composeRules= makeRuleTransducerFromFile(argv[3]);
    CompositeTransducer a4(&a3,composeRules);
    if (debuglev) cout << a4 << endl;

    // Word ending rules
    cout << "Rules 2" << endl;	
    RuleTransducer *wordEndRules= makeRuleTransducerFromFile(argv[4]);
    CompositeTransducer a5(&a4, wordEndRules);
    if (debuglev) cout << a5 << endl;

    cout << "Rules 3" << endl;
    RuleTransducer *lastRules= makeRuleTransducerFromFile(argv[5]);
    CompositeTransducer a6(&a5, lastRules);
    if (debuglev) cout << a6 << endl;
    
    cout << "stats a1 :" 
	 << a1.getMaxLabel()
	 << " noeuds. a3 " << a3.getMaxLabel() 
	 << " noeuds. a4 " << a4.getMaxLabel() 
	 << " noeuds. a5 " << a5.getMaxLabel()
	 << " noeuds. a6 " << a6.getMaxLabel()
	 << endl;

    //  a4.computeShortestPath();
    //      a4.showPath(cout, 1);
    //      cout << endl;
    //      a4.showPath(cout);

    cout << "==================================================" << endl;    
    a6.computeShortestPath();
    a6.showPath(cout, 1);
    cout << endl;
    a6.showPath(cout);
    cout << endl;
  } 
  catch (string e)
    {
      cerr << "l'erreur \"" << e << "\" s'est produite !" << endl;
    }
  catch (...)
    {
      cerr << "Une exception non identifie s'est produite" << endl;
    }
}


RuleTransducer *
makeRuleTransducerFromFile(const string &fname)
{
  char line[1024];
  RuleTransducer * a= new RuleTransducer();
  RuleNode *rn= a->getStart();
  rn->setFinal(true);
  
  ifstream s(fname.c_str());
  while (s.getline(line, 1024))
    {
      while (line[0] == '#' && s)
	{
	  line[0]= 0;
	  s.getline(line, 1024);
	}
      if (trace) cout << line << endl;
      try {
	rn->addLink(rn, line);
      } 
      catch (...) {
	cerr << "la ligne " << line << "est incorrecte" << endl;
	throw;
      }
    }
  s.close();
  return a;
}

RuleTransducer *
makeInsertSep()
{
  RuleTransducer *a= new RuleTransducer();
  
  RuleNode * start= a->getStart();
  RuleNode * rn0= a->makeNodeByLabel(3);
  RuleNode * rn1= a->makeNodeByLabel(1);
  RuleNode * rn2= a->makeNodeByLabel(2);
  start->addLink(rn0, "epsilon => debut");
  rn0->addLink(rn0, "debut  => debut");
  rn0->addLink(rn0, "fin => fin");
  rn0->addLink(rn1, "$A & ! fin => $A");
  rn1->addLink(rn1, "$A & ! fin => $A");
  rn1->addLink(rn1, "$A & ! fin => fin, debut, $A");
  rn1->addLink(rn0, "fin => fin");
  rn1->addLink(rn0, "debut => debut");
  rn1->addLink(rn2, "epsilon => fin");
  rn0->addLink(rn2, "epsilon => fin");
  rn2->setFinal(true);
  return a;
}


RuleTransducer *
makeWordCost()
{
  // Cost of words according to their length
  // The cost of a 1 consonant word : 200
  // The cost of a 2 consonant word : 100
  // The cost of a 3 consonant word : 0
  // The cost of a 4 consonant word : 200
  // The cost of a 5 consonant word : 400
  // More : + 400 / sign
  
  
  RuleTransducer *a= new RuleTransducer();
  
  RuleNode * rn0= a->getStart();
  rn0->setFinal(true);

  RuleNode * rn1= a->makeNodeByLabel(1);
  RuleNode * rn2= a->makeNodeByLabel(2);
  RuleNode * rn3= a->makeNodeByLabel(3);
  RuleNode * rn4= a->makeNodeByLabel(4);
  RuleNode * rn5= a->makeNodeByLabel(5);

  rn0->addLink(rn0, "debut=> epsilon)");
  rn0->addLink(rn0, "fin=> epsilon");
  rn0->addLink(rn0, "DET($X) => DET($X) / 100000");
  rn0->addLink(rn0, "R($X) => R($X)");
  rn0->addLink(rn0, "W($X), fin => W($X), fin");
  rn0->addLink(rn1, "L($X) => L($X)");
  rn0->addLink(rn1, "NUM($X), fin => NUM($X), fin");
  
  rn1->addLink(rn1, "DET($X) => DET($X)");
  rn1->addLink(rn1, "R($X) => R($X)");
  rn1->addLink(rn0, "fin => fin / 200");
  rn1->addLink(rn2, "L($X) => L($X)");

  rn2->addLink(rn2, "DET($X) => DET($X)");
  rn2->addLink(rn2, "R($X) => R($X)");
  rn2->addLink(rn0, "fin => fin / 100");
  rn2->addLink(rn3, "L($X) => L($X)");

  rn3->addLink(rn3, "DET($X) => DET($X)");
  rn3->addLink(rn3, "R($X) => R($X)");
  rn3->addLink(rn0, "fin => fin");
  rn3->addLink(rn4, "L($X) => L($X)");

  rn4->addLink(rn4, "DET($X) => DET($X)");
  rn4->addLink(rn4, "R($X) => R($X)");
  rn4->addLink(rn0, "fin => fin / 210");
  rn4->addLink(rn5, "L($X) => L($X) / 1600");

  rn5->addLink(rn5, "DET($X) => DET($X)");
  rn5->addLink(rn5, "R($X) => R($X)");
  rn5->addLink(rn0, "fin => fin");
  rn5->addLink(rn5, "L($X) => L($X) / 800");

  return a;
}
